@@ -1,5 +1,5 @@
 import os
-from torch_gauge import o3, verlet_list
+from torch_gauge.verlet_list import VerletList
 from torch_gauge._version import get_versions
 
 __version__ = get_versions()["version"]
