@@ -3,7 +3,7 @@ import math
 import torch
 import torch.nn.functional as F
 
-from torch_gauge.o3.spherical import SphericalTensor
+from torch_gauge.o3 import O3Tensor, SphericalTensor
 
 
 class SSP(torch.nn.Softplus):
@@ -42,7 +42,7 @@ class Swish(torch.nn.Module):
 
 
 class IELin(torch.nn.Module):
-    """
+    r"""
     Irrep-wise Equivariant Linear Layer.
 
     This module takes a spherical tensor and perform linear transformation within
@@ -70,31 +70,44 @@ class IELin(torch.nn.Module):
             for each l (and the associated m(s)).
         metadata_out (torch.LongTensor): the number of non-degenerate feature channels of the output tensor
             for each l (and the associated m(s)).
+        group (str): The group index of the tensors to be passed.
     """
 
-    def __init__(self, metadata_in, metadata_out):
+    def __init__(self, metadata_in, metadata_out, group="o3"):
         super().__init__()
         assert metadata_in.dim() == 1
         assert len(metadata_in) == len(metadata_out)
         self._metadata_in = metadata_in
         self._metadata_out = metadata_out
-        n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0)) * 2 + 1
+        group = group.lower()
+        if group == "so3":
+            self.tensor_class = SphericalTensor
+            self.n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0)) * 2 + 1
+        elif group == "o3":
+            self.tensor_class = O3Tensor
+            n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0) // 2) * 2 + 1
+            self.n_irreps_per_l = n_irreps_per_l.repeat_interleave(2)
+        else:
+            raise NotImplementedError(f"The group {group} is not supported in IELin")
         # The bias must be False
         self.linears = torch.nn.ModuleList(
             [
                 torch.nn.Linear(
-                    metadata_in[l] * n_irreps_per_l[l],
-                    metadata_out[l] * n_irreps_per_l[l],
+                    metadata_in[l],
+                    metadata_out[l],
                     bias=False,
                 )
-                if self._metadata_out[l] > 0
+                if self._metadata_out[l] > 0 and self._metadata_in[l] > 0
                 else None
                 for l, _ in enumerate(metadata_in)
             ]
         )
-        self._end_inds_in = torch.cumsum(self._metadata_in * n_irreps_per_l, dim=0)
+        self._end_inds_in = torch.cumsum(self._metadata_in * self.n_irreps_per_l, dim=0)
         self._start_inds_in = torch.cat([torch.LongTensor([0]), self._end_inds_in[:-1]])
-        self.out_layout = SphericalTensor.generate_rep_layout_1d_(self._metadata_out)
+        self.out_layout = torch.nn.Parameter(
+            self.tensor_class.generate_rep_layout_1d_(self._metadata_out),
+            requires_grad=False,
+        )
         self.num_out_channels = torch.sum(self._metadata_out).item()
 
     def forward(self, x: SphericalTensor) -> SphericalTensor:
@@ -114,18 +127,27 @@ class IELin(torch.nn.Module):
         outs = []
         for l, linear_l in enumerate(self.linears):
             if linear_l is None:
+                if self._metadata_out[l] > 0:
+                    outs.append(
+                        torch.zeros(
+                            *x.shape[:-1],
+                            self._metadata_out[l] * self.n_irreps_per_l[l],
+                            dtype=x.ten.dtype,
+                            device=x.ten.device,
+                        )
+                    )
                 continue
             in_l = x.ten[
                 ..., self._start_inds_in[l] : self._end_inds_in[l]
             ].contiguous()
-            out_l = linear_l(in_l)
-            outs.append(out_l)
+            out_l = linear_l(in_l.unflatten(-1, (self.n_irreps_per_l[l], self._metadata_in[l])))
+            outs.append(out_l.view(*x.shape[:-1], self._metadata_out[l] * self.n_irreps_per_l[l]))
 
         out_ten = torch.cat(outs, dim=-1)
         out_metadata = x.metadata.clone()
         out_metadata[-1] = self._metadata_out
-        out_rep_layout = x.rep_layout[:-1] + (self.out_layout,)
-        return SphericalTensor(
+        out_rep_layout = x.rep_layout[:-1] + (self.out_layout.data,)
+        return self.tensor_class(
             out_ten,
             rep_dims=x.rep_dims,
             metadata=out_metadata,
@@ -135,25 +157,32 @@ class IELin(torch.nn.Module):
 
 
 class RepNorm1d(torch.nn.Module):
-    """
+    r"""
     The (experimental) Representation Normalization layer.
 
     .. math::
 
-        \mathrm{RepNorm}(\mathbf{h})_{l,m} = \\frac{\mathrm{BatchNorm}(||\mathbf{h}_l||)}{\sqrt{2l+1}}
+        \mathrm{RepNorm}(\mathbf{h})_{l,m} = \mathrm{BatchNorm}(||\mathbf{h}_l||)
         \cdot \\frac{\mathbf{h}_{l}}{ |(1-{\\beta})||\mathbf{h}_{l}|| + {\\beta}| + \\epsilon}
 
     Heuristically, the trainable :math:`\mathbf{\\beta}` controls the fraction of norm
     information to be retained.
     """
 
-    def __init__(self, num_channels, momentum=0.1, eps=1e-2):
+    def __init__(self, num_channels, norm="batch", momentum=0.1, eps=0.1):
         super().__init__()
         self._num_channels = num_channels
         self._eps = eps
-        self.batchnorm = torch.nn.BatchNorm1d(
-            num_features=num_channels, momentum=momentum
-        )
+        if norm == "batch":
+            self.norm = torch.nn.BatchNorm1d(
+                num_features=num_channels, momentum=momentum, affine=False
+            )
+        elif norm == "node":
+            self.norm = torch.nn.LayerNorm(
+                normalized_shape=num_channels, elementwise_affine=False
+            )
+        else:
+            raise NotImplementedError
         # TODO: initialization schemes
         self.beta = torch.nn.Parameter(torch.rand(self._num_channels))
 
@@ -170,7 +199,8 @@ class RepNorm1d(torch.nn.Module):
         """
         x0 = x.invariant()
         assert x0.dim() == 2
-        x1 = self.batchnorm(x0)
+        x1 = self.norm(x0)
         divisor = torch.abs(x0.mul(1 - self.beta) + self.beta) + self._eps
         x2 = x.scalar_mul(1 / divisor)
+        # x2.ten = x2.ten.div((x.rep_layout[0][0] * 2 + 1).type(x0.dtype).sqrt().unsqueeze(0))
         return x1, x2
