@@ -18,7 +18,7 @@ from sympy import N
 from sympy.physics.quantum.cg import CG
 
 from torch_gauge import ROOT_DIR
-from torch_gauge.o3.spherical import SphericalTensor
+from torch_gauge.o3 import O3Tensor, SphericalTensor
 from torch_gauge.o3.wigner import csh_to_rsh
 
 memory = Memory(os.path.join(ROOT_DIR, ".o3_cache"), verbose=0)
@@ -185,7 +185,7 @@ class CGCoupler(torch.nn.Module):
         parity=0,
         overlap_out=True,
         trunc_in=True,
-        dtype=torch.float,
+        dtype=torch.double,
     ):
         super().__init__()
         metadata_1 = torch.LongTensor(metadata_1)
@@ -199,7 +199,10 @@ class CGCoupler(torch.nn.Module):
         self.parity = parity
         self.dtype = dtype
         self._init_params(overlap_out, trunc_in)
-        self.out_layout = SphericalTensor.generate_rep_layout_1d_(self.metadata_out)
+        self.out_layout = torch.nn.Parameter(
+            SphericalTensor.generate_rep_layout_1d_(self.metadata_out),
+            requires_grad=False,
+        )
 
     def _init_params(self, overlap_out, trunc_in):
         metadata_in = torch.stack([self.metadata_in1, self.metadata_in2], dim=0)
@@ -285,6 +288,7 @@ class CGCoupler(torch.nn.Module):
             repids_out.append(repids_out_3j)
             if not overlap_out:
                 out_ns_offset += degeneracy
+            lout_last = lout
 
         self.cg_tilde = torch.nn.Parameter(
             torch.cat(cg_tilde).type(self.dtype), requires_grad=False
@@ -342,3 +346,205 @@ class CGCoupler(torch.nn.Module):
             metadata=self.metadata_out.unsqueeze(0),
             rep_layout=(self.out_layout,),
         )
+
+
+class CGPCoupler(torch.nn.Module):
+    """
+    Parity-aware vectorized Clebsch-Gordan coupling module.
+
+    Note:
+        When the CGPCoupler class is instantiated, a compact view of representation indices
+        is generated for vectorizing Clebsch-Gordan coupling between two O3Tensors. Depending
+        on the setup of input O3Tensors, this tabulating step can be time-consuming; the CGPCoupler
+        parameters should be saved when the user intends to use the model for inference.
+
+    Attributes:
+        metadata_in1 (torch.LongTensor): The metadata of the first input O3Tensor to be coupled.
+        metadata_in2 (torch.LongTensor): The metadata of the second input O3Tensor to be coupled.
+        metadata_out (torch.LongTensor): The metadata of the output O3Tensor. Note that it depends
+            on the coupling specifications, `overlap_out` and `trunc_in`.
+
+    Args:
+        metadata_1 (torch.LongTensor): The representation metadata of the first tensor to be coupled.
+        metadata_2 (torch.LongTensor): The representation metadata of the second tensor to be coupled,
+             must have the same length (number of ls and ps) as ``metadata_1``.
+        trunc_in (bool): If true, the allowed feature indices (n) will be further truncated such that for
+             each set of terms (l1, l2, n), the coupling results will saturate all possible (l_out, n) values
+             of the output O3Tensor.
+        dtype (torch.dtype): The dtype for tensor to be passed in coupling, must be specified beforehand.
+
+    """
+
+    def __init__(
+        self,
+        metadata_1: torch.LongTensor,
+        metadata_2: torch.LongTensor,
+        trunc_in=True,
+        dtype=torch.double,
+    ):
+        super().__init__()
+        metadata_1 = torch.LongTensor(metadata_1)
+        metadata_2 = torch.LongTensor(metadata_2)
+        assert metadata_1.dim() == 1
+        assert metadata_2.dim() == 1
+        assert metadata_1.shape[0] == metadata_2.shape[0]
+        assert metadata_1.shape[0] % 2 == 0
+        self.metadata_out = None
+        self.metadata_in1 = metadata_1
+        self.metadata_in2 = metadata_2
+        self.dtype = dtype
+        self._init_params(trunc_in)
+        self.out_layout = torch.nn.Parameter(
+            O3Tensor.generate_rep_layout_1d_(self.metadata_out), requires_grad=False
+        )
+
+    def _init_params(self, trunc_in):
+        metadata_in = torch.stack([self.metadata_in1, self.metadata_in2], dim=0)
+        max_n_out = torch.maximum(self.metadata_in1, self.metadata_in2)
+        n_irreps_per_l = torch.arange(start=0, end=metadata_in.shape[1] // 2) * 2 + 1
+        n_irreps_per_lp = n_irreps_per_l.repeat_interleave(2)
+        repid_offsets_in = torch.cumsum(
+            metadata_in * n_irreps_per_lp.unsqueeze(0), dim=1
+        )
+        repid_offsets_in = torch.cat(
+            [torch.LongTensor([[0], [0]]), repid_offsets_in[:, :-1]], dim=1
+        ).long()
+        cg_tilde, repids_in1, repids_in2, repids_out = [], [], [], []
+        max_l = metadata_in.shape[1] // 2 - 1
+        # Tabulate the output metadata and allowed coupling terms
+        valid_coupling_ids = []
+        metadata_out = torch.zeros_like(max_n_out)
+        for lout in range(max_l + 1):
+            for pout in (1, -1):
+                for lin1 in range(max_l + 1):
+                    for lin2 in range(max_l + 1):
+                        for pin1 in (1, -1):
+                            for pin2 in (1, -1):
+                                coupling_parity = (-1) ** (lout + lin1 + lin2)
+                                # parity selection rule
+                                if pin1 * pin2 * coupling_parity != pout:
+                                    continue
+                                # Angular selection rule
+                                if lin1 + lin2 < lout or abs(lin1 - lin2) > lout:
+                                    continue
+                                lpin1 = 2 * lin1 + (1 - pin1) // 2
+                                lpin2 = 2 * lin2 + (1 - pin2) // 2
+                                lpout = 2 * lout + (1 - pout) // 2
+
+                                if trunc_in:
+                                    if lin1 + lin2 > max_l:
+                                        continue
+                                    degeneracy = min(
+                                        metadata_in[0, lpin1],
+                                        metadata_in[1, lpin2],
+                                        max_n_out[2 * (lin1 + lin2) + (1 - pout)//2],
+                                    )
+                                else:
+                                    if lout > max_l:
+                                        continue
+                                    degeneracy = min(
+                                        metadata_in[0, lpin1],
+                                        metadata_in[1, lpin2],
+                                        max_n_out[lpout],
+                                    )
+                                metadata_out[lpout] += degeneracy
+                                if degeneracy > 0:
+                                    valid_coupling_ids.append(
+                                        (lpout, lpin1, lpin2, degeneracy)
+                                    )
+
+        repid_offsets_out = torch.cumsum(metadata_out * n_irreps_per_lp, dim=0)
+        repid_offsets_out = torch.cat(
+            [torch.LongTensor([0]), repid_offsets_out[:-1]], dim=0
+        )
+
+        out_ns_offset, lpout_last = 0, 0
+        # Generate flattened coupling coefficients
+        for (lpout, lpin1, lpin2, degeneracy) in valid_coupling_ids:
+            if lpout > lpout_last:
+                out_ns_offset = 0
+            lin1, lin2, lout = lpin1 // 2, lpin2 // 2, lpout // 2
+            cg_source = get_rsh_cg_coefficients(lin1, lin2, lout)
+            cg_segment = cg_source.repeat_interleave(degeneracy, dim=1)
+            ns_segment = torch.arange(degeneracy).repeat(cg_source.shape[1])
+            # Calculating the representation IDs for the coupling tensors
+            repids_in1_3j = (
+                repid_offsets_in[0, lpin1]
+                + (cg_segment[0] + lin1) * metadata_in[0, lpin1]
+                + ns_segment
+            )
+            repids_in2_3j = (
+                repid_offsets_in[1, lpin2]
+                + (cg_segment[1] + lin2) * metadata_in[1, lpin2]
+                + ns_segment
+            )
+            repids_out_3j = (
+                repid_offsets_out[lpout]
+                + (cg_segment[2] + lout) * metadata_out[lpout]
+                + out_ns_offset
+                + ns_segment
+            )
+
+            cg_tilde.append(cg_segment[3])
+            repids_in1.append(repids_in1_3j)
+            repids_in2.append(repids_in2_3j)
+            repids_out.append(repids_out_3j)
+            out_ns_offset += degeneracy
+            lpout_last = lpout
+
+        self.cg_tilde = torch.nn.Parameter(
+            torch.cat(cg_tilde).type(self.dtype), requires_grad=False
+        )
+        self.repids_in1 = torch.nn.Parameter(
+            torch.cat(repids_in1).long(), requires_grad=False
+        )
+        self.repids_in2 = torch.nn.Parameter(
+            torch.cat(repids_in2).long(), requires_grad=False
+        )
+        self.repids_out = torch.nn.Parameter(
+            torch.cat(repids_out).long(), requires_grad=False
+        )
+        # Do not transfer metadata to device
+        self.metadata_out = metadata_out
+
+    def forward(self, x1: O3Tensor, x2: O3Tensor) -> O3Tensor:
+        """
+        Args:
+            x1 (O3Tensor): The first input ``O3Tensor`` to be coupled,
+                must have exactly 1 representation dimension.
+            x2 (O3Tensor): The second input ``O3Tensor`` to be coupled,
+                must have exactly 1 representation dimension.
+
+        Returns:
+            A new O3Tensor with ``self.metadata_out`` from C-G coupling.
+        """
+        assert len(x1.rep_dims) == 1
+        assert len(x2.rep_dims) == 1
+        assert x1.rep_dims[0] == x2.rep_dims[0]
+        coupling_dim = x1.rep_dims[0]
+        assert torch.all(x1.metadata[0].eq(self.metadata_in1))
+        assert torch.all(x2.metadata[0].eq(self.metadata_in2))
+        x1_tilde = torch.index_select(x1.ten, dim=coupling_dim, index=self.repids_in1)
+        x2_tilde = torch.index_select(x2.ten, dim=coupling_dim, index=self.repids_in2)
+        broadcast_shape = tuple(
+            self.cg_tilde.shape[0] if d == coupling_dim else 1
+            for d in range(x1.ten.dim())
+        )
+        out_tilde = x1_tilde * x2_tilde * self.cg_tilde.view(broadcast_shape)
+        out_shape = tuple(
+            self.out_layout.shape[1] if d == coupling_dim else x1.ten.shape[d]
+            for d in range(x1.ten.dim())
+        )
+        out_ten = torch.zeros(
+            out_shape, dtype=x1_tilde.dtype, device=x1_tilde.device
+        ).index_add_(
+            coupling_dim,
+            self.repids_out,
+            out_tilde,
+        )
+        return O3Tensor(
+            out_ten,
+            rep_dims=(coupling_dim,),
+            metadata=self.metadata_out.unsqueeze(0),
+            rep_layout=(self.out_layout.data,),
+        )
